# cube's specific schema
"""

:organization: Logilab
:copyright: 2009-2010 LOGILAB S.A. (Paris, FRANCE), license is LGPL v2.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: GNU Lesser General Public License, v2.1 - http://www.gnu.org/licenses
"""
from yams.buildobjs import EntityType, RelationDefinition, String, Interval
from cubicweb import _
from cubicweb.schema import ERQLExpression, RRQLExpression


class CWAuthCookie(EntityType):
    __permissions__ = {
        'read': ('managers', ERQLExpression('X owned_by U')),
        'add': ('managers', 'users'),
        'update': ('managers', 'owners'),
        'delete': ('managers', 'owners'),
        }

    magic = String(required=True, maxsize=128, unique=True,
                   description=_('unique identifier for the user'))
    lifetime = Interval()


class auth_cookie_for_user(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', RRQLExpression('U identity O', 'O')),
        'delete': ('managers', RRQLExpression('U identity O', 'O')),
        }

    subject = 'CWAuthCookie'
    object = 'CWUser'

    cardinality = '1*'
    composite = 'object'

    inlined = True



"""rememberme components: plug authentication retriever

:organization: Logilab
:copyright: 2009-2012 LOGILAB S.A. (Paris, FRANCE), license is LGPL v2.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: GNU Lesser General Public License, v2.1 - http://www.gnu.org/licenses
"""
from datetime import date, timedelta

from six import text_type

from cubicweb import Unauthorized, _
from cubicweb.utils import make_uid
from cubicweb.predicates import is_instance
from cubicweb.web import formfields as ff, formwidgets as fw
from cubicweb.web.views import (authentication, basecontrollers, basetemplates,
                                autoform, actions, uicfg)

_pvs = uicfg.primaryview_section
_pvs.tag_object_of(('CWAuthCookie', 'auth_cookie_for_user', 'CWUser'), 'hidden')

_abaa = uicfg.actionbox_appearsin_addmenu
_abaa.tag_object_of(('CWAuthCookie', 'auth_cookie_for_user', 'CWUser'), False)

# web authentication info retriever ############################################

class AuthCookieRetriever(authentication.WebAuthInfoRetriever):
    """authenticate by a magic identifier stored a cookie on the client host
    """
    __regid__ = 'magicnumauth'
    order = 0

    def __init__(self, vreg):
        # XXX make lifetime configurable
        self.cookie_lifetime = timedelta(days=30)

    def authentication_information(self, req):
        """retrieve authentication information from the given request, raise
        NoAuthInfo if expected information is not found:

        return info from long lived authentication cookie if any
        """
        cookie = req.get_cookie()
        try:
            authcookie = cookie['__cw_auth_cookie'].value
        except KeyError:
            raise authentication.NoAuthInfo()
        if not isinstance(authcookie, text_type):
            authcookie = authcookie.decode('ascii')
        try:
            login, magic = authcookie.split('|__magic=', 1)
        except ValueError:
            req.remove_cookie('__cw_auth_cookie')
            raise authentication.NoAuthInfo()
        return login, {'magic': magic}

    def cleanup_authentication_information(self, req):
        cookie = req.get_cookie()
        req.remove_cookie('__cw_auth_cookie')

    def authenticated(self, retriever, req, session, login, authinfo):
        """callback when return authentication information have opened a
        repository connection successfully. Take care req has no session
        attached yet, hence req.execute isn't available.

        Here we set authentication cookie for the next time
        """
        if not (retriever is self or '__setauthcookie' in req.form):
            return
        req.form.pop('__setauthcookie', None)
        user = session.user
        cls = req.vreg['etypes'].etype_class('CWAuthCookie')
        #  req has no session set yet
        with session.new_cnx() as cnx:
            authentity = cls.cw_instantiate(cnx.execute,
                                            magic=text_type(make_uid()),
                                            lifetime=self.cookie_lifetime,
                                            auth_cookie_for_user=user.eid)
            cnx.commit()
        # we've to commit here, else cookie may be rollbacked by errors while
        # trying to set last_login_time in the CookieSessionHandler (eg for ldap
        # user at least).
        magic = authentity.magic
        value = '%s|__magic=%s' % (user.login, magic)
        req.set_cookie('__cw_auth_cookie', value, maxage=None,
                       expires=date.today() + self.cookie_lifetime)
        authinfo['magic'] = magic

    def request_has_auth_info(self, req):
        cookie = req.get_cookie()
        try:
            authcookie = cookie['__cw_auth_cookie'].value
        except KeyError:
            return False
        try:
            login, magicnum = authcookie.split('|__magic=', 1)
        except ValueError:
            req.remove_cookie('__cw_auth_cookie')
            return False
        if not (login and magicnum):
            req.remove_cookie('__cw_auth_cookie')
            return False
        return True

    def revalidate_login(self, req):
        cookie = req.get_cookie()
        authcookie = cookie['__cw_auth_cookie'].value
        return authcookie.split('|__magic=', 1)[0]


# add remember me checkbox on the login form ###################################

basetemplates.LogForm.append_field(
    ff.BooleanField(name='__setauthcookie', choices=[(u'', u'1')],
                    label=_('remember me'), widget=fw.CheckBox({'class': 'data'}))
    )


# deactivate manual addition/edition ###########################################

class UneditableCWAuthCookieEdition(autoform.AutomaticEntityForm):
    __select__ = is_instance('CWAuthCookie')
    def __init__(self, *args, **kwargs):
        raise Unauthorized()

actions.ModifyAction.__select__ = actions.ModifyAction.__select__ & ~is_instance('CWAuthCookie')
actions.MultipleEditAction.__select__ = actions.MultipleEditAction.__select__ & ~is_instance('CWAuthCookie')



class LogoutController(basecontrollers.LogoutController):

    def publish(self, rset=None):
        """logout from the instance"""
        cookie = self._cw.get_cookie()
        if '__cw_auth_cookie' in cookie:
            self._cw.remove_cookie('__cw_auth_cookie')
        try:
            # removing all existing auth cookie entities, not only the currently
            # used one, doesn't seem like a bad idea
            self._cw.execute('DELETE CWAuthCookie X '
                             'WHERE X auth_cookie_for_user U, U eid %(u)s',
                             {'u': self._cw.user.eid})
            self._cw.cnx.commit()
        except:
            self.exception('error while deleting authentication cookie for %s',
                           self._cw.user.login)
        return super(LogoutController, self).publish(rset)


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, (LogoutController,))
    vreg.register_and_replace(LogoutController, basecontrollers.LogoutController)



"""http://fishbowl.pastiche.org/2004/01/19/persistent_login_cookie_best_practice/

:organization: Logilab
:copyright: 2009-2011 LOGILAB S.A. (Paris, FRANCE), license is LGPL v2.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: GNU Lesser General Public License, v2.1 - http://www.gnu.org/licenses
"""
__docformat__ = "restructuredtext en"

from cubicweb import AuthenticationError
from cubicweb.server.sources import native

class AuthCookieAuthentifier(native.BaseAuthentifier):
    auth_rql = ('Any X,AC WHERE X is CWUser, X login %(login)s, '
                'AC auth_cookie_for_user X, AC magic %(magic)s, '
                'AC lifetime L, AC creation_date > (NOW - L)')

    def authenticate(self, session, login, magic=None, **kwargs):
        """return CWUser eid for the given login/cookie if this account is
        defined in this source, else raise `AuthenticationError`
        """
        if magic is None:
            raise AuthenticationError()
        rset = session.execute(self.auth_rql, {'login': login,
                                               'magic': magic})
        if rset:
            session.repo.glob_delete_entities(session, set((rset[0][1],)))
            session.commit(free_cnxset=False)
            return rset[0][0]
        raise AuthenticationError('bad cookie')




"""remberme cube hooks

:organization: Logilab
:copyright: 2009-2011 LOGILAB S.A. (Paris, FRANCE), license is LGPL v2.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: GNU Lesser General Public License, v2.1 - http://www.gnu.org/licenses
"""
__docformat__ = "restructuredtext en"

from cubicweb.server import hook

from cubes.rememberme.authplugin import AuthCookieAuthentifier

class ServerStartupHook(hook.Hook):
    """task to cleanup expirated auth cookie entities"""
    __regid__ = 'authcookieinit'
    events = ('server_startup',)

    def __call__(self):
        # XXX use named args and inner functions to avoid referencing globals
        # which may cause reloading pb
        def cleanup_auth_cookies(repo=self.repo):
            with repo.internal_cnx() as cnx:
                cnx.execute('DELETE CWAuthCookie X WHERE X lifetime L, '
                            'X creation_date < (NOW - L)')
                cnx.commit()
        self.repo.looping_task(60*60, cleanup_auth_cookies, self.repo)
        self.repo.system_source.add_authentifier(AuthCookieAuthentifier())



from datetime import timedelta

from cubicweb.devtools.webtest import CubicWebTestTC


class RememberMeAuthenticationTC(CubicWebTestTC):
    def init_auth_cookie(self, login, magic):
        self.auth_cookie.value = '%s|__magic=%s' % (login, str(magic))
        self.webapp.cookiejar.set_cookie(self.auth_cookie)

    def init_login_password(self, login):
        req, origcnx = self.init_authentication('cookie')
        req.session = req.cnx = None
        req.form['__login'] = login
        req.form['__password'] = login
        return req, origcnx

    def test_auth_cookie_authentication(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, 'user')

        # valid login / password, not asked to remember user
        self.login('user')
        self.assertEqual('user', self.webapp.cookies['__data_session'][:4])

        with self.admin_access.repo_cnx() as cnx:
            cookies = cnx.find(
                'CWUser', login='user'
            ).one().reverse_auth_cookie_for_user
            self.assertEqual(len(cookies), 0)

        self.logout()
        self.assertNotIn('__data_session', self.webapp.cookies)

        # valid login / password, ask to remember user
        self.login('user', __setauthcookie=1)
        self.assertEqual('user', self.webapp.cookies['__data_session'][:4])

        for cookie in self.webapp.cookiejar:
            if cookie.name == '__cw_auth_cookie':
                self.auth_cookie = cookie

        with self.admin_access.repo_cnx() as cnx:
            cookies = cnx.find(
                'CWUser', login='user'
            ).one().reverse_auth_cookie_for_user
            self.assertEqual(len(cookies), 1)

            magic = cookies[0].magic
            self.assertEqual(cookies[0].lifetime, timedelta(days=30))

        self.webapp.reset()

        # bad magic number
        self.init_auth_cookie('user', '1234')
        self.webapp.get('/')
        self.assertEqual('anon', self.webapp.cookies['__data_session'][:4])

        # bad login in auth cookie
        self.init_auth_cookie('admin', magic)
        self.webapp.get('/')
        self.assertEqual('anon', self.webapp.cookies['__data_session'][:4])

        # valid auth cookie
        self.init_auth_cookie('user', magic)
        self.webapp.get('/')
        self.assertEqual('user', self.webapp.cookies['__data_session'][:4])

        with self.admin_access.repo_cnx() as cnx:
            cookies = cnx.find(
                'CWUser', login='user'
            ).one().reverse_auth_cookie_for_user
            self.assertEqual(len(cookies), 1)

            magic2 = cookies[0].magic
            self.assertNotEqual(magic2, magic)

        self.logout()


if __name__ == '__main__':
    import unittest
    unittest.main()



"""

:organization: Logilab
:copyright: 2001-2009 LOGILAB S.A. (Paris, FRANCE), license is LGPL v2.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
:license: GNU Lesser General Public License, v2.1 - http://www.gnu.org/licenses
"""
import os
import pwd

from logilab.common.pytest import PyTester

def getlogin():
    """avoid usinng os.getlogin() because of strange tty / stdin problems
    (man 3 getlogin)
    Another solution would be to use $LOGNAME, $USER or $USERNAME
    """
    return pwd.getpwuid(os.getuid())[0]


def update_parser(parser):
    login = getlogin()
    parser.add_option('-r', '--rebuild-database', dest='rebuild_db',
                      default=False, action="store_true",
                      help="remove tmpdb and rebuilds the test database")
    parser.add_option('-u', '--dbuser', dest='dbuser', action='store',
                      default=login, help="database user")
    parser.add_option('-w', '--dbpassword', dest='dbpassword', action='store',
                      default=login, help="database name")
    parser.add_option('-n', '--dbname', dest='dbname', action='store',
                      default=None, help="database name")
    parser.add_option('--euser', dest='euser', action='store',
                      default=login, help="esuer name")
    parser.add_option('--epassword', dest='epassword', action='store',
                      default=login, help="euser's password' name")
    return parser


class CustomPyTester(PyTester):
    def __init__(self, cvg, options):
        super(CustomPyTester, self).__init__(cvg, options)
        if options.rebuild_db:
            os.unlink('tmpdb')
            os.unlink('tmpdb-template')



# pylint: disable=W0622
# copyright 2013-2018 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-rememberme application packaging information"""

from os import listdir as _listdir
from os.path import join, isdir
from glob import glob

modname = 'rememberme'
distname = 'cubicweb-rememberme'

numversion = (1, 1, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
description = 'implementation of the Remember Me feature for the CubicWeb framework'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ = {
    'cubicweb': '>= 3.23.0',
}
__recommends__ = {}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
    ]


THIS_CUBE_DIR = join('share', 'cubicweb', 'cubes', modname)


def listdir(dirpath):
    return [join(dirpath, fname) for fname in _listdir(dirpath)
            if fname[0] != '.' and not fname.endswith('.pyc')
            and not fname.endswith('~')
            and not isdir(join(dirpath, fname))]


data_files = [
    # common files
    [THIS_CUBE_DIR, [fname for fname in glob('*.py') if fname != 'setup.py']],
    ]

# check for possible extended cube layout
for dname in ('entities', 'views', 'sobjects', 'hooks',
              'schema', 'data', 'wdoc', 'i18n', 'migration', 'test'):
    if isdir(dname):
        data_files.append([join(THIS_CUBE_DIR, dname), listdir(dname)])
# Note: here, you'll need to add subdirectories if you want
# them to be included in the debian package
